# -*- ruby -*-
# $Id: umail.rb,v 1.8 2003/06/24 11:24:28 tommy Exp $

require 'nkf'

class UMail

  @@strictly = false

  def UMail::strictly=(flag)
    @@strictly = flag
  end

  def UMail::strictly()
    @@strictly
  end

  module Base64
    module_function
    # B󥳡ɤ줿إåʸǥɤ
    def hdecode(str)
      str.unpack('m')[0]
    end
    # Base64 ǥ
    def decode(str)
      Base64::hdecode str
    end
  end

  module QuotedPrintable
    module_function
    # Q󥳡ɤ줿إåʸǥɤ
    def QuotedPrintable::hdecode(str)
      str.gsub('_', ' ').gsub(/=([0-9A-F][0-9A-F])/o) do $1.hex.chr end
    end
    # QuotedPrintable ǥ
    def QuotedPrintable::decode(str)
      str.gsub(/[ \t]+$/o, '').gsub(/=\r?\n/o, '').
	gsub(/=([0-9A-F][0-9A-F])/o) do $1.hex.chr end
    end
  end

  @@nkfFlag = {'iso-2022-jp' => 'J', 'euc-jp' => 'E', 'shift_jis' => 'S'}

  # ĤΥȡǥɤ
  def UMail::mdecode_token(s)
    if s !~ /\A=\?([a-z0-9_-]+)\?(Q|B)\?([^?]+)\?=\Z/io then
      s
    else
      charset, encoding, text = $1, $2, $3
      c = @@nkfFlag[charset.downcase]
      if c == nil then return s end
      if encoding.downcase == 'q' then s = QuotedPrintable.hdecode(text)
      else s = Base64.hdecode(text)
      end
      NKF::nkf("-e#{c}m0", s)
    end
  end

  # ʣΥȡ()ǥɤ
  def UMail::mdecode(list, strictly=nil)
    strictly = UMail::strictly? if strictly == nil
    if not strictly then
      list.join.gsub(/\?=\s+=\?/, '?==?').gsub(/=\?[a-z0-9_-]+\?(Q|B)\?[^?]+\?=/io) do UMail::mdecode_token($&) end
    else
      cs = ''
      sp = ''
      after_encode = false
      list.each do |c|
	c = c.to_s
	if c =~ /\A\s+\Z/ then c = ' ' end
	if after_encode and c == ' ' then sp = c; next end
	c2 = UMail::mdecode_token c
	if after_encode and c != c2 then cs << c2; next end
	after_encode = c != c2
	cs << sp + c2
	sp = ''
      end
      cs
    end
  end

  class ParseError < StandardError
  end

  class FWS
    def inspect() "FWS" end
    def to_s() " " end
  end
  FWS0 = FWS::new

  class CFWS < FWS
    def initialize(str)
      @comment = str
    end
    def inspect() "CFWS" end
    def to_s() @comment end
    attr_reader :comment
  end

  class TokenList < Array

    def [](*args)
      ret = super
      ret.type == Array ? TokenList[ret] : ret
    end

    # Ϣ³Ĥˤ
    def compact_ws!()
      ret = TokenList::new
      self.each do |m|
	if m.is_a? FWS and ret.last.is_a? FWS then
	  next
	end
	ret << m == CFWS ? UMail::FWS0 : m
      end
      self.replace ret
      ret
    end

    # 
    def delete_ws!()
      self.replace self.map{|m| m.is_a?(FWS) ? nil : m}.compact
    end

    # Ƭζ
    def skip_ws!()
      while self[0].is_a? FWS do
	shift
      end
      self
    end

    #  d ʬ䤹
    def split(d)
      ret = []
      s = self.dup
      loop do
	i = s.index(d)
	if i == nil then
	  ret << s
	  break
	end
	ret << s.slice!(0,i+1)
	ret.last.pop
      end
      if block_given? then
	ret.each do |i| yield i end
	return nil
      end
      ret
    end

  end


  # ԤΥإå
  class Header

    def initialize(s)
      @raw_body = s
      @strictly = UMail::strictly
    end

    def to_s()
      @raw_body
    end

    def body()
      parse unless @parsed
      @parsed_body
    end

    def strictly?()
      @strictly
    end

    attr_accessor :strictly

    private

    def parse()
      @parsed_body = @raw_body
      @parsed = true
    end

    def mdecode(list)
      UMail::mdecode list, strictly?
    end

  end

  # ʸإå
  class StringH < Header

    private

    def parse()
      @parsed_body = mdecode(@raw_body.split(/(\s+)/o))
      @parsed = true
    end
  end

  # ¤إå
  class StructH < Header

    WDay = ['Mon','Tue','Wed','Thu','Fri','Sat','Sun']
    Mon = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec']
    ZoneStricted = {
      'UT' => 0, 'GMT' => 0,
      'EST' => -5, 'EDT' => -4,
      'CST' => -6, 'CDT' => -5,
      'MST' => -7, 'MDT' => -6,
      'PST' => -8, 'PDT' => -7,
      'A' => 1, 'B' => 2, 'C' => 3, 'D' => 4, 'E' => 5, 'F' => 6,
      'G' => 7, 'H' => 8, 'I' => 9, 'K' => 10, 'L' => 11, 'M' => 12,
      'N' => -1, 'O' => -2, 'P' => -3, 'Q' => -4, 'R' => -5, 'S' => -6,
      'T' => -7, 'U' => -8, 'V' => -9, 'W' => -10, 'X' => -11, 'Y' => -12,
      'Z'=>0, 
    }

    Zone = ZoneStricted.dup
    Zone['JST'] = 9
    
    Specials = "()<>@,;:\\\".[]".split(//)
    TokenDelimiterRe = /[][()<>@,;:\\.]|\s+|\"(?:[^\\\"]|\\.)*\"/

    # Ȥ
    def comments()
      parse unless @parsed
      @comments
    end

    def inspect()
      "#<#{type}:#{body}>" rescue "#<#{type}:#{to_s}>"
    end

    private

    # 
    def parse()
      @tokens = split_token @raw_body
      @parsed_body = @tokens.join
      @comments = @tokens.map{|t| t.type == CFWS ? t.to_s : nil}.compact
      @parsed = true
    end

    # ʸ s ȡʬ䤹
    def split_token(s, re=TokenDelimiterRe)
      s0 = s
      ret = TokenList::new
      while s =~ re
	ret << $` if $` != ''
	m, s = $&, $'
	case m[0]
	when ?(
	  cnt = 1
	  com = [m]
	  while cnt > 0
	    if s == '' then raise ParseError::new "')' not found: #{s0}" end
	    if s =~ /\A\s+|\A([^()\\\s]|\\.)*/o then
	      com << $&
	      s = $'
	    end
	    case s[0]
	    when ?)
	      com << s.slice!(0,1)
	      cnt -= 1
	    when ?(
	      com << s.slice!(0,1)
	      cnt += 1
	    end
	  end
	  com.delete ''
	  ret << CFWS::new(mdecode(com))
	when 0x20, 0x09
	  ret << UMail::FWS0 unless ret[-1] == UMail::FWS0
	else
	  ret << m
	end
      end
      ret << s
      ret.delete ''
      if ret[0] == ' ' then ret.shift end
      if ret[-1] == ' ' then ret.pop end
      ret
    end

  end

  # շإå
  class DateH < StructH

    def date()
      parse unless @parsed
      @date
    end

    private

    # 
    def parse()
      super
      r = @tokens
      r.delete_ws!
      if WDay.include? r[0] then
	if r[1] != ',' then raise ParseError::new "comma required: #{@raw_body}" end
	r[0,2] = nil
      end
      if r.length < 6 then raise ParseError::new "invalid format: #{@raw_body}" end
      unless r[0] =~ /^\d+$/o and Mon.include? r[1] and r[2] =~ /^\d+$/o and
	  r[3] =~ /^\d+$/o and r[4] == ':' and r[5] =~ /^\d+$/o then
	raise ParseError::new "invalid format: #{@raw_body}"
      end
      day, mon, year, hour, min = r[0].to_i, Mon.index(r[1])+1, r[2].to_i, r[3].to_i, r[5].to_i
      r[0,6] = nil
      sec = 0
      if r[0] == ':' then
	if r.length < 2 then raise ParseError::new "invalid format: #{@raw_body}" end
	sec = r[1].to_i
	r[0,2] = nil
      end
      if 70 <= year and year < 100 then year += 1900 end
      if 70 > year then year += 2000 end
      zone = r.shift
      if zone == nil and strictly? then raise ParseError::new "timezone required: #{@raw_body}" end

      t = Time::gm(year, mon, day, hour, min, sec)
      zl = strictly? ? ZoneStricted : Zone
      if zl.include? zone then
	t -= zl[zone] * 60 * 60
      elsif zone =~ /^(\+|-)(\d\d)(\d\d)$/o then
	if $1 == '+' then
	  t -= $2.to_i * 60 * 60 + $3.to_i * 60
	else
	  t += $2.to_i * 60 * 60 + $3.to_i * 60
	end
      end
      @date = t.localtime
    end

  end

  class RecvH < StructH
  end

  class RetpathH < StructH
  end

  # ᡼륢ɥ쥹
  class MailAddr

    def initialize(addr, phrase=nil, comments=nil)
      @addr = addr
      @phrase = phrase
      @comments = comments
    end

    attr_reader :addr, :phrase, :comments

    def to_s()
      @phrase ? "#{@phrase} <#{@addr}>" : "<#{@addr}>"
    end

    def inspect()
      "#<#{type}:#{to_s}>"
    end

  end

  class MailGroup

    def initialize(name, addrs)
      @name = name
      @addrs = addrs
    end

    attr_reader :name, :addrs

    def to_s()
      if name == nil then
	addrs.map{|a| a.to_s}.join(", ")
      elsif addrs.empty? then
	"#{@name}:;"
      else
	"#{@name}: "+addrs.map{|a| a.to_s}.join(", ")+";"
      end
    end

    def inspect()
      "#<#{type}:#{to_s}>"
    end

  end

  class XaddrH < StructH

    private

    def parse_addr(tokens)
      comments = tokens.map{|t| t.type == CFWS ? t.to_s : nil}.compact
      tokens.compact_ws!
      tokens.skip_ws!
      if i = tokens.index('<') then
	phrase = tokens[0 .. i-1] if i > 0
	if j = tokens.index('>') and i < j then
	  addr = tokens[i+1 .. j-1]
	  tokens[0 .. j] = nil
	else
	  raise ParseError::new "invalid '<': #{tokens.join}"
	end
      else
	phrase = nil
	addr = tokens.dup
	tokens.clear
      end
      tokens.skip_ws!
      if not tokens.empty? then
	raise ParseError::new "invalid format: #{tokens.join}"
      end
      if phrase then
	phrase.skip_ws!
	if phrase[-1].is_a? FWS then phrase.pop end
	if strictly? then
	  phrase.each do |p|
	    if Specials.include? p then
	      raise ParseError::new "invalid phrase '#{p}' in '#{phrase.join}'"
	    end
	  end
	end
	phrase = mdecode phrase
      end
      addr.delete_ws!
      if strictly? then
	if addr.empty? then
	  raise ParseError::new "empty address"
	end
	if not addr.include? '@' then
	  raise ParseError::new "'@' not in address: #{tokens.join}"
	end
      end
      MailAddr::new(addr.join, phrase, comments.empty? ? nil : comments)
    end
  end

  # ᡼륢ɥ쥹إå
  class SaddrH < XaddrH

    def addr()
      parse unless @parsed
      @addr
    end

    private

    def parse()
      super
      @addr = parse_addr @tokens
      @parsed_body = @addr.to_s
    end

  end

  # ʣ᡼륢ɥ쥹إå
  class MaddrH < XaddrH

    def addrs()
      parse unless @parsed
      @addrs
    end

    def addr()
      addrs[0]
    end

    private

    def parse()
      super
      @addrs = []
      @tokens.split(",") do |a|
	@addrs << parse_addr(a)
      end
      @parsed_body = @addrs.map{|a| a.to_s}.join(', ')
    end

  end

  # 롼׻ǽ᡼륢ɥ쥹إå
  class GaddrH < XaddrH

    def groups()
      parse unless @parsed
      @groups
    end

    def addrs()
      parse unless @parsed
      @addrs
    end

    def addr()
      addrs[0]
    end

    private

    def parse()
      super
      @groups = []
      @addrs = []
      tokens = @tokens
      tokens.compact_ws!
      tokens.skip_ws!
      loop do
	i = tokens.index(":")
	k = tokens.index(",")
	if i and (not k or (k and i < k)) then
	  groupname = tokens.slice!(0, i+1)
	  groupname.pop
	  addrs = []
	  tokens.skip_ws!
	  if not j = tokens.index(';') then
	    raise ParseError::new "invalid group format: #{tokens.join}"
	  end
	  if strictly? then
	    groupname.each do |p|
	      if Specials.include? p then
		raise ParseError::new "invalid group name '#{p}' in '#{groupname.join}'"
	      end
	    end
	  end
	  groupname = mdecode groupname
	  mlist = tokens.slice!(0, j+1)
	  mlist.pop
	  if not mlist.empty? then
	    mlist.split(",") do |a|
	      addr = parse_addr(a)
	      addrs << addr
	      @addrs << addr
	    end
	  end
	  @groups << MailGroup::new(groupname, addrs)
	  tokens.skip_ws!
	  break if tokens.empty?
	  if tokens[0] != "," then
	    raise ParseError::new "invalid format: #{tokens.join}"
	  end
	  tokens.shift
	  tokens.skip_ws!
	elsif k then
	  addr = parse_addr(tokens.slice!(0,k))
	  @addrs << addr
	  tokens.shift
	  @groups << MailGroup::new(nil, [addr])
	else
	  addr = parse_addr tokens
	  @addrs << addr
	  @groups << MailGroup::new(nil, [addr])
	  break
	end
      end
      @parsed_body = @groups.map{|g| g.to_s}.join(', ')
    end

  end

  # åIDإå
  class MsgidH < StructH

    def msgid()
      parse unless @parsed
      @msgid
    end

    private

    def parse()
      super
      @tokens.delete_if{|t| t.is_a? FWS}
      if @tokens[0] != '<' or @tokens[-1] != '>' then
	raise ParseError::new "invalid format: #{@raw_body}"
      end
      @msgid = @tokens.join
    end

  end

  # ե󥹷إå
  class RefH < StructH

    def msgids()
      parse unless @parsed
      @msgids
    end

    def phrases()
      parse unless @parsed
      @phrases
    end

    private

    def parse()
      super
      @msgids = []
      @phrases = ['']
      in_msgid = nil
      @tokens.each_index do |i|
	case @tokens[i]
	when '<'
	  in_msgid = i
	  @phrases << ''
	when '>'
	  if not in_msgid then raise ParseError::new "invalid format: #{@raw_body}" end
	  id = @tokens[in_msgid..i]
	  id.delete_if{|t| t.is_a? FWS}
	  @msgids << id.join
	  in_msgid = nil
	else
	  if not in_msgid then
	    @phrases[-1] << @tokens[i].to_s
	  end
	end
      end
      @phrases.each do |i| i.strip! end
      if @phrases[-1] == '' then
	@phrases.pop
      end
      @parsed = true
    end

  end

  class KeyH < StructH
  end

  class EncH < StructH
  end

  class VersionH < StructH
  end

  # Content-Type إå
  class CTypeH < StructH

    def maintype()
      parse unless @parsed
      @maintype
    end

    def subtype()
      parse unless @parsed
      @subtype
    end

    def params()
      parse unless @parsed
      @params
    end

    private

    def parse()
      tokens = split_token @raw_body, /[][()<>@,;:\\\/=\?]|\s+|\"(?:[^\\\"]|\\.)*\"/o
      @parsed_body = tokens.join
      tokens.delete_if{|t| t.is_a? FWS}
      @maintype = tokens[0].downcase
      @subtype = nil
      @params = {}
      if tokens.length > 1 then
	if tokens[1] != '/' then raise ParseError::new "invalid format: #{@raw_body}" end
	@subtype = tokens[2].downcase
	tokens[0,3] = nil
	while tokens.length > 0
	  if tokens[0] != ';' or tokens[2] != '=' then
	    raise ParseError::new "invalid format: #{@raw_body}" if strictly?
	    break
	  end
	  pn, pv = tokens[1], tokens[3]
	  @params[pn.downcase] = pv.gsub(/^\"|\"$/, '')
	  tokens[0,4] = nil
	end
      end
      @parsed = true
    end

  end

  # Content-Transfer-Encoding إå
  class CEncodingH < StructH
    def encoding()
      parse unless @parsed
      @encoding
    end

    private

    def parse()
      super
      @tokens.delete_if{|t| t.is_a? FWS}
      if strictly? and @tokens.length != 1 then
	raise ParseError::new "invalid format: #{@raw_body}"
      end
      @encoding = @tokens[0].downcase
    end
  end

  # Content-Disposition إå
  class CDispositionH < StructH

    def disposition()
      parse unless @parsed
      @disposition
    end

    def params()
      parse unless @parsed
      @params
    end

    private

    def parse()
      tokens = split_token @raw_body, /[][()<>@,;:\\\/=\?]|\s+|\"(?:[^\\\"]|\\.)*\"/o
      @parsed_body = tokens.join
      tokens.delete_if{|t| t.is_a? FWS}
      @disposition = tokens[0].downcase
      tokens[0,1] = nil
      @params = {}
      while tokens.length > 0
	if tokens[0] != ';' or tokens[2] != '=' then
	  raise ParseError::new "invalid format: #{@raw_body}" if strictly?
	  break
	end
	pn, pv = tokens[1], tokens[3]
	@params[pn.downcase] = pv.gsub(/^\"|\"$/, '')
	tokens[0,4] = nil
      end
      @parsed = true
    end

  end

  HeaderClass = {
    'date'                      => DateH,
    'resent-date'               => DateH,
    'received'                  => RecvH,
    'return-path'               => RetpathH,
    'sender'                    => SaddrH,
    'resent-sender'             => SaddrH,
    'to'                        => GaddrH,
    'cc'                        => GaddrH,
    'bcc'                       => GaddrH,
    'from'                      => MaddrH,
    'reply-to'                  => GaddrH,
    'resent-to'                 => GaddrH,
    'resent-cc'                 => GaddrH,
    'resent-bcc'                => GaddrH,
    'resent-from'               => MaddrH,
    'resent-reply-to'           => GaddrH,
    'message-id'                => MsgidH,
    'resent-message-id'         => MsgidH,
    'content-id'                => MsgidH,
    'in-reply-to'               => RefH,
    'references'                => RefH,
    'keywords'                  => KeyH,
    'encrypted'                 => EncH,
    'mime-version'              => VersionH,
    'content-type'              => CTypeH,
    'content-transfer-encoding' => CEncodingH,
    'content-disposition'       => CDispositionH,
    'subject'                   => StringH,
    'comments'                  => StringH,
    'content-description'       => StringH
  }

  def initialize(str)
    @strictly = @@strictly
    @header = {}
    @raw = str
    @header_s, @body_s = @raw.gsub(/\r\n/o, "\n").split(/\n\n/o, 2)
    @header_s.gsub(/\n[ \t]+/o, " ").split(/\n/o).each do |s|
      hname, hbody = s.split(/:\s*/o, 2)
      hn = hname.downcase
      c = HeaderClass.has_key?(hn) ? HeaderClass[hn] : StringH
      @header[hn] = [] if not @header.has_key? hn
      @header[hn] << c::new(hbody)
    end
  end

  attr_reader :header

  def [](n)
    @header[n.downcase]
  end

  def each()
    @header.each do |a|
      yield a
    end
  end

  def has_key?(k)
    @header.has_key?(k)
  end

  def from()
    if @header.has_key? "from" and @header["from"][0].addrs.length > 0 then
      @header["from"][0].addrs[0].to_s
    end
  end

  def from_addr()
    if @header.has_key? "from" and @header["from"][0].addrs.length > 0 then
      @header["from"][0].addrs[0].addr
    else
      nil
    end
  end

  def from_phrase()
    if @header.has_key? "from" and @header["from"][0].addrs.length > 0 then
      @header["from"][0].addrs[0].phrase
    else
      nil
    end
  end

  def to()
    if @header.has_key? "to" then
      @header["to"].map{|a| a.addrs.map{|b| b.to_s}}.flatten
    else
      nil
    end
  end

  def to_addr()
    if @header.has_key? "to" then
      @header["to"].map{|a| a.addrs.map{|b| b.addr}}.flatten
    else
      nil
    end
  end

  def to_phrase()
    if @header.has_key? "to" then
      @header["to"].map{|a| a.addrs.map{|b| b.phrase}}.flatten
    else
      nil
    end
  end

  def cc()
    if @header.has_key? "cc" then
      @header["cc"].map{|a| a.addrs.map{|b| b.to_s}}.flatten
    else
      nil
    end
  end

  def cc_addr()
    if @header.has_key? "cc" then
      @header["cc"].map{|a| a.addrs.map{|b| b.addr}}.flatten
    else
      nil
    end
  end

  def cc_phrase()
    if @header.has_key? "cc" then
      @header["cc"].map{|a| a.addrs.map{|b| b.phrase}}.flatten
    else
      nil
    end
  end

  def date()
    if @header.has_key? "date" then
      @header["date"][0].date
    else
      nil
    end
  end

  def subject()
    if @header.has_key? "subject" then
      @header["subject"][0].body
    else
      nil
    end
  end

  def maintype()
    if @header.has_key? "content-type" then
      @header["content-type"][0].maintype
    else
      nil
    end
  end

  def subtype()
    if @header.has_key? "content-type" then
      @header["content-type"][0].subtype
    else
      nil
    end
  end

  def charset()
    if @header.has_key? "content-type" and @header["content-type"][0].params.has_key? "charset" then
      @header["content-type"][0].params["charset"]
    else
      nil
    end
  end

  def filename()
    fn = nil
    if @header.has_key? "content-type" and @header["content-type"][0].params.has_key? "name" then
      fn = @header["content-type"][0].params["name"]
    elsif @header.has_key? "content-disposition" and @header["content-disposition"][0].params.has_key? "filename" then
      fn = @header["content-disposition"][0].params["filename"]
    end
    fn ? UMail::mdecode(fn.split(/(\s+)/o), strictly) : nil
  end

  def multipart?()
    maintype == "multipart"
  end

  def body()
    parse_body if not @body_parsed
    @body
  end

  def body_umail()
    parse_body if not @body_parsed
    @body_umail
  end

  def preamble()
    parse_body if not @body_parsed
    @preamble
  end

  def epilogue()
    parse_body if not @body_parsed
    @epilogue
  end

  def parts()
    parse_body if not @body_parsed
    @parts_a
  end

  def to_s()
    @raw
  end

  def inspect()
    "#<#{type}>"
  end

  attr_accessor :strictly

  private

  def parse_body()
    if not multipart? then
      @body = @body_s
      if @header.has_key? "content-transfer-encoding" then
	case @header["content-transfer-encoding"][0].encoding
	when "base64"
	  @body = Base64::decode(@body)
	when "quoted-printable"
	  @body = QuotedPrintable::decode(@body)
	end
      end
      if @header.has_key? "content-type" then
	if maintype == "text" then
	  if @header["content-type"][0].params.has_key? "charset" then
	    case @header["content-type"][0].params["charset"].downcase
	    when "iso-2022-jp"
	      @body = NKF::nkf("-Jem0", @body)
	    when "shift_jis"
	      @body = NKF::nkf("-Sem0", @body)
	    end
	  else
	    @body = NKF::nkf("-Jem0", @body)
	  end
	end
      else
	@body = NKF::nkf("-Jem0", @body)
      end
    else
      b = @header["content-type"][0].params["boundary"]
      if b == nil then raise ParseError::new "no boundary in content-type" end
      bre = Regexp::new '^'+Regexp::quote("--#{b}")+'[ \t]*\n'
      cre = Regexp::new '^'+Regexp::quote("--#{b}--")+'[ \t]*(\n|\Z)'
      a = @body_s.split(bre)
      if a[-1] =~ cre then
	a[-1] = $`	#`
	epi = $'	#'
      else
	raise ParseError::new 'no close-delimiter'
      end
      @body = @preamble = a.shift
      @epilogue = epi
      @parts_a = a.map{|m| UMail::new m}
      if @parts_a[0].maintype == "text" then
	@body_umail = @parts_a.shift
	@body = @body_umail.body
      end
    end
    @body_parsed = true
  end

end
